@@ -1098,7 +1098,12 @@ glib_protocol_version_from_gnutls (gnutls_protocol_t protocol_version)
 static gchar *
 get_ciphersuite_name (gnutls_session_t session)
 {
+#if GTLS_GNUTLS_CHECK_VERSION(3, 7, 4)
+  return g_strdup (gnutls_ciphersuite_get (session));
+#else
   gnutls_protocol_t protocol_version = gnutls_protocol_get_version (session);
+  char *cipher_name;
+  char *result;
 
   if (protocol_version <= GNUTLS_TLS1_2 ||
       (protocol_version >= GNUTLS_DTLS0_9 && protocol_version <= GNUTLS_DTLS1_2))
@@ -1108,9 +1113,20 @@ get_ciphersuite_name (gnutls_session_t session)
                                                      gnutls_mac_get (session)));
     }
 
-  return g_strdup_printf ("TLS_%s_%s",
-                          gnutls_cipher_get_name (gnutls_cipher_get (session)),
-                          gnutls_digest_get_name (gnutls_prf_hash_get (session)));
+  cipher_name = g_strdup (gnutls_cipher_get_name (gnutls_cipher_get (session)));
+  for (char *c = cipher_name; *c != '\0'; c++)
+    {
+      if (*c == '-')
+        *c = '_';
+    }
+
+  result = g_strdup_printf ("TLS_%s_%s",
+                            cipher_name,
+                            gnutls_digest_get_name (gnutls_prf_hash_get (session)));
+  g_free (cipher_name);
+
+  return result;
+#endif
 }
 
 static void
@@ -1160,6 +1176,16 @@ gnutls_get_binding (GTlsConnectionGnutls      *gnutls,
 
   if (ret == GNUTLS_E_SUCCESS)
     {
+      /* Older GnuTLS versions are known to return SUCCESS and empty data for TLSv1.3 tls-unique binding.
+       * While it may look prudent to catch here that specific corner case, the empty binding data is
+       * definitely not a SUCCESS, regardless of the version and type. */
+      if (cb.size == 0)
+        {
+          g_set_error (error, G_TLS_CHANNEL_BINDING_ERROR, G_TLS_CHANNEL_BINDING_ERROR_GENERAL_ERROR,
+                       _("Empty channel binding data indicates a bug in the TLS library implementation"));
+          return FALSE;
+        }
+
       if (data != NULL)
         {
           g_tls_log_debug (gnutls, "binding size %d", cb.size);
